# Whodunit: CodeChef AI & Human Solutions Dataset

This repository contains the data for the [CodeChef](https://www.codechef.com/) problems and the code used to collect, extract code-style and code complexity features from it, as well as the code to build and evaluate the classifiers for the paper `Whodunit: Classifying Code as Human Authored or GPT-4 Generated - A case study on CodeChef problems`. It also contains the modified baseline code.

## Data Collection
 The data corresponds to `399` problems filtered from the initial set of `1100` problems. The code used for collection and filtering can be found in the scripts directory. The final data is stored as described below.

### Files

- `final_dataset.json`: Contains the data for all the problems on CodeChef.
- `final_successful_dataset.json`: Contains the same data as `final_dataset.json` but the `ai_solutions` field only contains the solutions that successfully passed the public test cases.

The `final_files` directory contains zipped sets of python files (binned into `easy`, `medium` and `hard` difficulties) that were used for the experiments in the paper.
- `unverified.zip`: Contains 2 Human and 2 AI solutions for each problem.
- `verified.zip`: Contains solutions that are verified to pass the public test cases.
- `sampled.zip`: Contains a random set of problems sources from the `unverified` set such that it matches the distribution of the `verified` set.


### Data Format
The JSON data is stored as a nested dictionary. The top level keys are the `11 difficulty levels` of CodeChef. Each difficulty level is a dictionary with the key being the `problem_code_id` on CodeChef and the value being the data for the problem. The data for each problem is structured in a dictionary with the following keys:

- `constraints`: A string describing any constraints related to the problem.
- `subtasks`: A string detailing the subtasks associated with the problem.
- `sample_test_cases`: An array of dictionaries, each representing a public test case. Each test case includes:
    - `input`: The input given to the problem.
    - `output`: The expected output for the given input.
    - `explanation`: A detailed explanation of why the output is as expected.
- `problem_statement`: A string describing the problem, its background, and requirements.
- `input_format`: A string describing the format in which input is provided.
- `output_format`: A string describing the format in which output is expected.
- `problem_name`: The name of the problem.
- `user_tags`: An array of strings representing user-defined tags for the problem.
- `computed_tags`: An array of strings representing system-generated tags for the problem.
- `problem_code_id`: A string representing the unique code ID of the problem.
- `difficulty_level`: A string or number indicating the difficulty level of the problem.
- `ai_solutions`: An array of strings, each representing GPT-4 (v0613) generated solution to the problem.
- `human_solutions`: An array of dictionaries, each containing details about a solution submitted by a user, which includes:
    - `id`: A unique identifier for the solution.
    - `submission_date`: The date of submission.
    - `language`: The programming language used.
    - `username`: The username of the submitter.
    - `user_rating_star`: The user's rating.
    - `contest_code`: The code of the contest in which the solution was submitted.
    - `tooltip`: Status of the solution (e.g., accepted, rejected).
    - `score`: The score achieved by the solution.
    - `points`: The points achieved by the solution.
    - `icon`: A link to an icon representing the status of the solution.
    - `time`: The execution time of the solution.
    - `memory`: The memory used by the solution.
    - `solution`: A unique identifier for the solution.
    - `code`: The actual code of the solution.


**Note:** The `input_format`, `output_format` and `constraints` fields are not available for older problems on CodeChef. In such cases, the information is present in the `problem_statement` field.


## Feature Extraction
Before extracting the features, the comments and multi-line strings are to be removed by running the `scripts/remove_all_comments.py` script. This script accepts the `source_directory`, `destination_directory` and `output_file_path` which are the paths to the directory containing the files, the directory to store the files with comments removed and the path to a file that logs information about the file and removal process.

The code for extracting the code-style and code complexity features from the dataset can be found in the `script/feature_extraction` directory. This directory has 3 files: `extract_main_features.py`, `extract_non_gameable_features.ipynb`, and `extract_with_halstead_features.ipynb` corresponding to the 3 different feature sets used in the paper. These scripts read the files from the input path provided and outputs the extracted features to the output path provided. The extracted features are stored in the `data/features` directory from which 8 features files are generated. The features files are named as follows:
- `main_features.csv`: Contains the main classifier's features (`RQ1`).
- `with_halstead_features.csv`: Contains the halstead features (`RQ1`).
- `non_gameable_features.csv`: Contains the non-gameable features (`RQ2`).
- `correct_solutions_features.csv`: Contains the features for solutions that passed the public test cases (`RQ3`).
- `sampled_solutions_features.csv`: Contains the features for solutions sampled from the unverified set (`RQ3`).
- `easy_problems_features.csv`: Contains the features for solutions to the easy problems (`RQ4`).
- `medium_problems_features.csv`: Contains the features for solutions to the intermediate problems (`RQ4`).
- `hard_problems_features.csv`: Contains the features for solutions to the hard problems (`RQ4`).

**Note:** The extraction scripts require that the directory to the respective dataset is provided as input and the output path is specified. The output path in this package is the `data/features` directory.

## Classification
There are 8 scripts corresponding to the 8 different classifiers used in the paper and the 8 extracted features files. The scripts are found in the `scripts/classification` directory. The scripts are named as follows:
- `main_classification.py`: Contains the code for the main classifier (`RQ1`).
- `with_halstead_classification.py`: Contains the code for the halstead classifier (`RQ1`).
- `non_gameable_classification.py`: Contains the code for the non-gameable classifier (`RQ2`).
- `correct_solutions_classification.py`: Contains the code for the classifier trained on solutions that passed the public test cases (`RQ3`).
- `sampled_solutions_classification.py`: Contains the code for the classifier trained on solutions sampled from the unverified set (`RQ3`).
- `easy_problems_classification.py`: Contains the code for the classifier trained on solutions to the easy problems (`RQ4`).
- `medium_problems_classification.py`: Contains the code for the classifier trained on solutions to the intermediate problems (`RQ4`).
- `hard_problems_classification.py`: Contains the code for the classifier trained on solutions to the hard problems (`RQ4`).

Each of these scripts accept a `.csv` file containing the features as input and prints out the results. The scripts also display the SHAP plot for the top 10 features.

## Baseline
The baseline code was retrieved from GitHub and can be found [here](https://github.com/ldklab/scored23_release) and modified to work on our dataset as best as we could. The modified code comprise several scripts and the steps taken to run them were taken from the GitHub repository. 

### Prerequisites
1. Download required packages using pip install -r requirements.txt
2. Clone and build "TreeSitter" package using the following steps
```bash
cd ASTanalysis
mdkir vendor && cd vendor
git clone https://github.com/tree-sitter/tree-sitter-cpp
cd ..
python builder.py 
# my-languages.so* file will be created under the folder ASTanalysis/build
```
3. Create a new folder "Processed" in the same directory as the "ASTAnalysis" folder.
4. Download and install MongoDB local instance. docker run -d -p 127.0.0.1:27017:27017 -v <path-to>/db:/data/db mongo:6
5. Download the ASTAnalysis folder. This contains the code files and the "Dataset" folder.

### Files
#### 1. ASTParser.py
This file parses each C/C++ code files available in the "Dataset" folder in to its Abstract Syntax Trees and saves it in a XML format to a destination folder and in a collection in MongoDB database. 

#### 2. ASTLooper.py
This file executes the ASTParser.py file that loops through the dataset.

#### 3. CreateNodeTypeSet.py
This will create pickle files with all the extracted node types from the ASTs for each code file in the dataset folder. It will also create a text file with all the node types found in all the code files in dataset folder.

#### 4. NGramEmptyPair.py
This file creates the empty dictionaries for bi,tri and quad grams and save them to a pickle file at the desired destination folder.

#### 5. DictOfNodes.py
This file creates a dictionary with the "Key" values as all the node types found within the code and values "Values" of those keys are the generalized node types. You can change the generalized node types from within this file. 

#### 6. BigramDictUpdate.py & TrigramDictUpdate.py & QuadgramDictUpdate.py
This will update the empty dictionary for each code file in source dataset based on the n-gram value (2,3,

#### 7. ASTandLexicalLooper.py 
This file loops through the source dataset files and executes the ASTandLexicalParse.py 

#### 8. ASTandLexicalParse.py
This file parses the source code files in the syntactic and lexical features and save those vectors in MongoDB. 

#### 9. ASTandLexicalClassification.py
This file applies the stratified binary classification using RFC, KNN, XGB & SVM on the created normalized vectors for syntactic and lexical features. 


### Follow Below Steps For Classification Based on Syntactic + Lexical Features
#### 9. ASTandLexParse.py

- Edit the MongoDB database details as per your own setup. The default settings are applied for local default MongoDB database instance.

#### 10. ASTandLexLooper.py

- This will execute the ASTandLexParse.py file. 

- Run the the file by providing two absolute paths of the below as inputs:

    1. Absolute path to the Dataset folder
    2. Absolute path to AST dictionary folder, either Bigram, Trigram or Quadgram that was created in step # 6, 7 & 8 

- To run the file: 

        python3 ASTandLexLooper.py /absolute/path/of/dataset/folder absolute/path/of/dictionary

- Based on the dictionary input file it will create a new Collection in MongoDB. Collection will be named as ```CombinedVectorWith<Bigram/Trigram/Quadgram>```

- Run the file again for N-gram length 3 and 4

#### 11. ASTandLexClassification.py

- Uncomment the appropriate line from line# 27, 28 or 29 to retrieve the vectors from the required MongoDB collection. 

        # mycol = mydb["CombinedVectorWithBigram"]
        # mycol = mydb["CombinedVectorWithTrigram"]
        # mycol = mydb["CombinedVectorWithQuadgram"]

- To run the file:

        python3 ASTandLexClassification.py

#### External Tools:
1. TreeSitter
2. NetworkX
3. MongoDB


## Other Files
- `scripts/requirements.txt`: Contains the list of packages required to run the feature extraction and classification scripts.
- `Online_Appendix.pdf`: Contains the appendix of the Python node types and keywords extracted as discussed in section`3.2` of the paper.